/* manStr.c
 * source file for the managed strings library
 */

/* defines */

#define put(format) (printf("[manStr]" format "\n"))
#define putf(format, ...) (printf("[manStr]" format "\n", __VA_ARGS__))

/* standard/lib includes */

#include<stdio.h>
#include<stdlib.h>
#include<stdarg.h>

/* local includes */

#include"manStr.h"

/* various cstr functions */

int cstr_pstrlen(char* p_src) {
	for(int i = 0;;i++) {
		if(p_src[i] == 0) return i;
	}
}

void cstr_pstrcpyp(char** p_target, char* p_src) {
	for(int i = 0; p_src[i] != 0; i++) {
		*p_target = realloc(*p_target, i + 2);
		(*p_target)[i] = p_src[i];
		(*p_target)[i + 1] = 0;
	}
}

/* managed strings */

int check(mstr_t p_in) {
	if (p_in.src[p_in.len] == 0) {
		return 1;
	} else {
		put("check failed");
		return 0;
	}
}

void clearString(mstr_t* p_str) {
	p_str->len = 0;
	if(p_str->src != NULL)
		free(p_str->src);
	p_str->src = calloc(1, 1);
}

mstr_t createStringU(char* p_src) {
	if(p_src[0] == 0 || p_src == 0) {
		return (mstr_t) {calloc(1, 1), 0};
	}
	mstr_t r_str = {NULL, cstr_pstrlen(p_src)};
	r_str.src = malloc(cstr_pstrlen(p_src));
	cstr_pstrcpyp(&r_str.src, p_src);
	return r_str;
}

mstr_t createString(char* p_src, unsigned int nul_loc) {
	if(p_src == NULL) {
		return (mstr_t) {0, 0};
	}

	if(p_src[nul_loc] != 0 && cstr_pstrlen(p_src) != nul_loc) {
		put("create string fucked up");
		return (mstr_t){0, 0};
	}
	
	return createStringU(p_src);
}

void resizeString(mstr_t* p_target, int size) {
	p_target->src = realloc(p_target->src, size + 1);
	p_target->len = size;
	p_target->src[size] = 0;
}

void rrsizeString(mstr_t* p_target, int rel) {
	if(p_target->len + rel < 0)
		return;
	p_target->len += rel;
	p_target->src = realloc(p_target->src, p_target->len + 1);
	p_target->src[p_target->len] = 0;
}

void insertcString(mstr_t* p_target, char c, int i) {
	rrsizeString(p_target, 1);
	for(int q = p_target->len; q > i; q--) {
		p_target->src[q] = p_target->src[q-1];
		p_target->src[q-1] = 0;
	}
	p_target->src[i] = c;
	p_target->src[p_target->len] = 0;
}

void copyString(mstr_t* p_target, mstr_t p_src) {
	if(p_src.len == 0) return;
	if(p_target->src != NULL) free(p_target->src);
	p_target->len = p_src.len;
	p_target->src = calloc(p_src.len, 1);
	for(int i = 0; i <= p_src.len; i++)
		p_target->src[i] = p_src.src[i];
}

void copyoString(mstr_t* p_target, mstr_t p_src, int b, int e) {
	if(p_target->src == NULL) {
		*p_target = mstr("");
	}

	if(e > 0)
		for(int i = b; i <= e && i <= p_src.src[i]; i++) {
			p_target->len++;
			p_target->src = realloc(p_target->src, p_target->len + 1);
			p_target->src[p_target->len - 1] = \
						str(p_src)[i];
			p_target->src[p_target->len] = 0;
		}
	else
		for(int i = b; i <= p_src.src[i]; i++) {
			p_target->len++;
			p_target->src = realloc(p_target->src, p_target->len + 1);
			p_target->src[p_target->len - 1] = \
						str(p_src)[i];
			p_target->src[p_target->len] = 0;
		
		}
}
void appendString(mstr_t* p_target, mstr_t p_src)
{
	for(int i = 0; i < p_src.len; i++)
	{
		appendcString(p_target, p_src.src[i]);
	}
}

void delString(mstr_t* p_target, int b, int e) {
	if(p_target->src == NULL) return;
	mstr_t tmp = mstr("");
	if(e == -1) e = p_target->len;
	for(int i = 0; i < b; i++)
		appendcString(&tmp, p_target->src[i]);
	for(int i = e + 1; i < p_target->len; i++)
		appendcString(&tmp, p_target->src[i]);
	copyString(p_target, tmp);
	p_target->len = tmp.len;
	freeString(&tmp);
}

mstr_t rappendString(mstr_t p_target, mstr_t p_src) {
	mstr_t buf = createStringU("");
	appendString(&buf, p_target);
	for(int i = buf.len; str(p_src)[i] != 0; i++) {
		buf.len++;
		buf.src = realloc(buf.src, i + 2);
		buf.src[i] = str(p_src)[i];
		buf.src[i + 1] = 0;
	}
	return buf;
}

void appendcString(mstr_t* p_target, char p_src) {
	p_target->len++;
	p_target->src = realloc(p_target->src, p_target->len + 1);
	p_target->src[p_target->len - 1] = p_src;
	p_target->src[p_target->len] = 0;
}

void astrf(mstr_t* p_target, const char* fmt, ...) {
	va_list args, args2;

	va_start(args, fmt);

	va_copy(args2, args);
	int prevLen = p_target->len;

	int q = vsnprintf(NULL, 0, fmt, args);
	p_target->len += q;
	p_target->src = realloc(p_target->src, p_target->len + 1);
	for(int i = prevLen; i <= prevLen + q; i++)
		p_target->src[i] = 0;

	vsnprintf(str(*p_target) + prevLen, q + 1, fmt, args2);

	va_end(args);
}

int compareString(mstr_t* p_target, mstr_t* p_src) {
	int total = 0;
	for(int i = 0; i < p_target->len && i < p_target->len; i++) {
		if(p_target->src[i] != p_src->src[i]) {
			total++;
		}
	}
	return total;
}

void freeString(mstr_t* p_target) {
	p_target->len = -1;
	if(p_target->src != NULL) {
		free(p_target->src);
	}
}
